{*******************************************************}
{                                                       }
{       GDICACHE 2.1.0.0                                }
{       ----------------                                }
{       ADV Compiled, original name "JeeFonts"          }
{       Copyright (C) 2000, Lucian Radulescu            }
{       E-mail     : ljee@interlog.com                  }
{       Home Page  : http://www.interlog.com/~ljee      }
{                                                       }
{*******************************************************}

unit GDIObjs;

{$I ALLLIB.DEF}

interface

uses
  Windows, SysUtils, Classes, Objects;

type

  { TGraphObject : generic graphical object. It will hold font, brush or pen }
  TGraphObject = class
  private
    FData: Pointer;                     { a TLogFont, TLogBrush or TLogPen }
    FSize: Word;                        { size of FData }
    procedure CreateHandle;             { Creates gdi object: font, brush, pen}
    procedure SetData(Value: Pointer);  { save gdi object creation data }
  protected
    FHandle: THandle;                   { gdi object handle }
    constructor Create(ASize: Word; AData: Pointer);
  public
    destructor Destroy; override;
    property Data: Pointer read FData write SetData;
  end;

  { TGraphObjList : sorted list of TGraphObject }
  TGraphObjList = class(TSafeSortedCollection)
  private
    FDC             : HDC;              { current output DC }
    FCurrent        : Pointer;          { current selected object }
    FLastData       : Pointer;          { last selected object FData }
    FLastIndexUsed  : Longint;          { indexof last selected object }
    SysGDIObject    : THandle;          { handle to the original graphic object }
    LastGDIObject   : THandle;          { temp handle of last gdi object used }
    procedure SetData(AData: Pointer);
  protected
    function NewItemAdd(AData: Pointer): TObject; virtual; abstract;
  public
    constructor Create(ADC: HDC); {$IFDEF ATLEAST_D4}reintroduce;{$ENDIF}
    destructor Destroy; override;
    function KeyOf(Item: Pointer): Pointer; override;
    property Current: Pointer read FCurrent write SetData;
    property LastData: Pointer read FLastData write FLastData;
  end;

  { TFontObject }
  TFontObject = class(TGraphObject)
  protected
    constructor Create(ALogFont: PLogFont);
  end;

  { TBrushObject }
  TBrushObject = class(TGraphObject)
  protected
    constructor Create(ALogBrush: PLogBrush);
  end;

  { TPenObject }
  TPenObject = class(TGraphObject)
  protected
    constructor Create(ALogPen: PLogPen);
  end;

  { TFontList }
  TFontList = class(TGraphObjList)
  protected
    function NewItemAdd(AData: Pointer): TObject; override;
  public
    function Compare(Key1, Key2: Pointer): LongInt; override;
  end;

  { TBrushList }
  TBrushList = class(TGraphObjList)
  protected
    function NewItemAdd(AData: Pointer): TObject; override;
  public
    function Compare(Key1, Key2: Pointer): LongInt; override;
  end;

  { TPenList }
  TPenList = class(TGraphObjList)
  protected
    function NewItemAdd(AData: Pointer): TObject; override;
  public
    function Compare(Key1, Key2: Pointer): LongInt; override;
  end;

procedure FreeGraph_Objects;

const
  FontList  : TFontList   = nil;
  BrushList : TBrushList  = nil;
  PenList   : TPenList    = nil;

implementation

procedure DisplaySysError;
var
  ErrorCode : Integer;
  Buf       : array [Byte] of Char;
begin
  ErrorCode := GetLastError;
  if (ErrorCode <> 0) and
     (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, ErrorCode,
		    LOCALE_USER_DEFAULT, Buf, sizeof(Buf), nil) <> 0) then
    raise Exception.Create(Buf)
  else
    raise Exception.Create('Out of resources');
end;

{.$WARNINGS OFF}
function GDICheck(Value: Cardinal): Cardinal;
begin
  if Value = 0 then DisplaySysError
  else Result := Value;
end;
{.$WARNINGS ON}

{ TGraphObject }
constructor TGraphObject.Create(ASize: Word; AData: Pointer);
begin
  inherited Create;
  FHandle := 0;
  FData   := nil;
  FSize   := ASize;
  Data    := AData;
end;

procedure TGraphObject.SetData(Value: Pointer);
begin
  GetMem(FData, FSize);
  Move(Value^, FData^, FSize);
end;

destructor TGraphObject.Destroy;
begin
  FreeMem(FData, FSize);
  if FHandle <> 0 then     {we automatically}
    DeleteObject(FHandle); {delete graphic objects}
  inherited Destroy;
end;

procedure TGraphObject.CreateHandle;
begin
  if Self is TFontObject then
    FHandle := GDICheck(CreateFontIndirect(TLogFont(FData^))) else
  if Self is TBrushObject then
    FHandle := GDICheck(CreateBrushIndirect(TLogBrush(FData^))) else
  if Self is TPenObject then
    FHandle := GDICheck(CreatePenIndirect(TLogPen(FData^)));
end;

{ TGraphObjList }
constructor TGraphObjList.Create(ADC: HDC);
begin
  inherited Create(10, 4);
  FDC            := ADC;
  FCurrent       := nil;
  FLastData      := nil;
  FLastIndexUsed := -1;
  SysGDIObject   := 0;
end;

destructor TGraphObjList.Destroy;
begin
  LastGDIObject := 0;
  if SysGDIObject <> 0 then
  begin
    LastGDIObject := SelectObject(FDC, SysGDIObject); {select back original}
    SysGDIObject  := 0;
  end;
  inherited Destroy;
end;

function TGraphObjList.KeyOf(Item: Pointer): Pointer;
begin
  Result := TGraphObject(Item).FData;
end;

{WHEN CACHING, handles of graphic objects are kept (and freed when we free the list)}
{WHEN NOT CACHING, handles of graphic objects are deleted here and recreated, based}
{on information stored in FData. When we free the list, we also must delete the last
{selected object.}
procedure TGraphObjList.SetData(AData: Pointer);
var
  OldSel: THandle;
begin
  EnterCS;
  try
    if not _UNSAFE_SearchKey(AData, FLastIndexUsed) then  {not in list ?}
      _UNSAFE_AtInsert(FLastIndexUsed, NewItemAdd(AData));{create+insert}
    FLastData := FCurrent;
    with TGraphObject(Items[FLastIndexUsed]) do           {use this object}
    begin
      LastGDIObject := SysGDIObject;                      {temp save}
      OldSel := SelectObject(FDC, FHandle);               {old selected handle}
      if SysGDIObject = 0 then SysGDIObject := OldSel;    {save original handle}
      FCurrent := FData;
    end;
  finally
    LeaveCS;
  end;
end;

{ TFontObject }
constructor TFontObject.Create(ALogFont: PLogFont);
begin
  inherited Create(SizeOf(TLogFont), ALogFont);
  CreateHandle;
end;

{ TBrushObject }
constructor TBrushObject.Create(ALogBrush: PLogBrush);
begin
  inherited Create(SizeOf(TLogBrush), ALogBrush);
  CreateHandle;
end;

{ TPenObject }
constructor TPenObject.Create(ALogPen: PLogPen);
begin
  inherited Create(SizeOf(TLogPen), ALogPen);
  CreateHandle;
end;

{ TFontList }
function TFontList.Compare(Key1, Key2: Pointer): LongInt;
begin
  with TLogFont(Key1^) do begin
    Result := StrIComp(lfFaceName, TLogFont(Key2^).lfFaceName);
    if Result = 0 then begin
      Result := ABS(lfHeight) - ABS(TLogFont(Key2^).lfHeight);
      if Result = 0 then begin
	Result := lfWeight - TLogFont(Key2^).lfWeight;
	if Result = 0 then begin
	  Result := lfWidth - TLogFont(Key2^).lfWidth;
	  if Result = 0 then begin
	    Result := lfItalic - TLogFont(Key2^).lfItalic;
	    if Result = 0 then begin
	      Result := lfEscapement - TLogFont(Key2^).lfEscapement;
	      if Result = 0 then begin
		Result := lfUnderline - TLogFont(Key2^).lfUnderline;
    	        if Result = 0 then
  		  Result := lfStrikeOut - TLogFont(Key2^).lfStrikeOut;
              end;
	    end;
	  end;
	end;
      end;
    end;
  end;
end;

function TFontList.NewItemAdd(AData: Pointer): TObject;
begin
  Result := TFontObject.Create(AData);
end;

{ TBrushList }
function TBrushList.Compare(Key1, Key2: Pointer): LongInt;
begin
  with TLogBrush(Key1^) do begin
    Result := lbStyle - TLogBrush(Key2^).lbStyle;
    if Result = 0 then begin
      Result := lbColor - TLogBrush(Key2^).lbColor;
      if Result = 0 then
	Result := lbHatch - TLogBrush(Key2^).lbHatch;
    end;
  end;
end;

function TBrushList.NewItemAdd(AData: Pointer): TObject;
begin
  Result := TBrushObject.Create(AData);
end;

{ TPenList }
function TPenList.Compare(Key1, Key2: Pointer): LongInt;
begin
  with TLogPen(Key1^) do begin
    Result := lopnStyle - TLogPen(Key2^).lopnStyle;
    if Result = 0 then begin
      Result := lopnWidth.X - TLogPen(Key2^).lopnWidth.X;
      if Result = 0 then
	Result := lopnColor - TLogPen(Key2^).lopnColor;
    end;
  end;
end;

function TPenList.NewItemAdd(AData: Pointer): TObject;
begin
  Result := TPenObject.Create(AData);
end;

procedure FreeGraph_Objects;
begin
  if Assigned(FontList) then FontList.Free;
  FontList := nil;
  if Assigned(BrushList) then BrushList.Free;
  BrushList := nil;
  if Assigned(PenList) then PenList.Free;
  PenList := nil;
end;

initialization
  ;
finalization
  FreeGraph_Objects;
end.

