// ThreadSafe Collections
//
// LIFO usage : Add - Last
// FIFO usage : Add - First

{$I ALLLIB.DEF}

unit Objects;

interface

uses
  Windows, SysUtils, Classes;

const
  HINDEX_ERROR =    -1;

type

  ECOLLECTION       = class(Exception);
  ELongFreeError    = class(ECOLLECTION)
    constructor Create;
  end;
  ELongRangeError   = class(ECOLLECTION)
    constructor Create;
  end;
  EDuplicates       = class(ECOLLECTION)
    constructor Create;
  end;

  ELongRealloc      = class(ECOLLECTION)
    constructor Create;
    constructor CreateInfo(ExceptInfo: String);
  end;
  ELongOverflow   = class(ELongRealloc)
    constructor Create;
  end;

  TKeyOfFunc        = function (Item: Pointer): Pointer;
  TCompareFunc      = function (Key1, Key2: Pointer): LongInt;
  TCollectionNotify = (cnAdded,   // notify item that it was inserted
                       cnAccessed,// notify item that it was accessed
                       cnDeleted, // notify item that it was removed
                       cnFree);   // notify item that it WILL be disposed

  // TSafeCollection : dynamic array of unsorted items
  TSafeCollection = class(TObject)
  private
    FCount   : Longint;            // number if items
    FDelta   : Longint;            // number of items to grow when required
    FSize    : Longint;            // number of allocated items (FSize can be > FCount)
    FMem     : Pointer;            // pointer to global memory
    FLock    : TRTLCriticalSection;// handle to critical section
    FForwrd  : Boolean;            // used only by ForEach to establish direction
    FCurr    : Longint;            // Updated by FirstThat, LastThat, NextThat.
                                   // When they finish, no need to call IndexOf,
                                   // use directly FCurr (public Curr)
    FUserCanAbort: Boolean;

    function  At(Index: Longint): Pointer;
    function  AtInsert(Index: Longint; Item: Pointer): Longint;
    function  CheckIndex(Index: Longint): Boolean;
    function  CheckSize(Value: Longint): Boolean;
    procedure CheckNotify(Item: Pointer; Action: TCollectionNotify);
    procedure DisposeItem(Item: Pointer);
    function  GetPtr(Index: Longint): PLongint;
    procedure MoveItems(Src, Dest: Longint);
    procedure SetItem(Index: Longint; Item: Pointer);
    procedure SetSize(NewSize: Longint);

  protected

    // If items are not objects, you MUST OVERRIDE FreeItem
    // and dispose data in your code.
    procedure FreeItem(Item: Pointer); virtual;

    // item notification
    procedure Notify(Item: Pointer; Action: TCollectionNotify); virtual;

    // swap two items
    procedure Swap(Index1, Index2: LongInt); virtual;
  public
    // constructor, destructor
    constructor Create(ArraySize, ADelta: Longint); virtual;
    destructor Destroy; override;

    // critical section protection
    procedure EnterCS;
    procedure LeaveCS;

    // Delete item(s) methods - ITEM(S) NOT DISPOSED
    // Deleted item(s) will be notified through Notify,
    // with Action set to cnDeleted.
    procedure Delete(Index: Longint);
    procedure DeleteItem(Item: Pointer);
    procedure DeleteAll;
    procedure Remove(Item: Pointer);

    // free item(s) methods - ITEM(S) IS DISPOSED through FreeItem.
    procedure Clear;
    procedure FreeAllItems;
    procedure FreeAtIndex(Index: Longint);
    procedure FreeThis(Item: Pointer); virtual;

    function  _UNSAFE_AtInsert(Index: Longint; Item: Pointer): Longint;
    procedure _Exchange(Index1, Index2: LongInt);
    procedure _Slide(Index1, Index2: LongInt); 

    function  Add(Item: Pointer): Longint;
    function  Append(Item: Pointer): Longint;

    function  Insert(Item: Pointer; KeyOfFunc: TKeyOfFunc;
                     CompareFunc: TCompareFunc): Longint; virtual;
    function  InsertItem(Item: Pointer): Longint; virtual;
    function  IndexOf(Item: Pointer; KeyOfFunc: TKeyOfFunc;
                      CompareFunc: TCompareFunc): LongInt; virtual;
    function  IndexOfItem(Item: Pointer): LongInt; virtual;
    function  Nth(Index : LongInt): Pointer;

    function  First: Pointer;
    function  Last: Pointer;
    function  Head: Pointer; 
    function  Tail: Pointer;

    procedure PlaceBefore(aItem, P: Pointer);

    function  FirstThat(Test: Pointer): Pointer;
    function  NextThat(Test: Pointer): Pointer;
    function  LastThat(Test: Pointer): Pointer;
    procedure ForEach(Action: Pointer);

    procedure Exchange(Index1, Index2: LongInt); virtual;
    procedure Slide(Index1, Index2: LongInt); virtual;

    property  Count : Longint read FCount;
    property  Curr  : Longint read FCurr write FCurr;
    property  Delta : Longint read FDelta write FDelta;
    property  Size  : Longint read FSize  write SetSize;
    property  UserCanAbort: Boolean read FUserCanAbort write FUserCanAbort;
    property  Items[Index: Longint]: Pointer read At write SetItem; default;
  end;

  // TSafeSortedCollection
  TSafeSortedCollection = class(TSafeCollection)
  private

    FDuplicates: Boolean;
    FSorted: Boolean;

  protected
    function _UNSAFE_Search(Key: Pointer;
                            var Index: LongInt;
                            KeyOfFunc: TKeyOfFunc;
                            CompareFunc: TCompareFunc): Boolean;
    function _UNSAFE_SearchKey(Key: Pointer; var Index: LongInt): Boolean;
    function Search(Key: Pointer;
                            var Index: LongInt;
                            KeyOfFunc: TKeyOfFunc;
                            CompareFunc: TCompareFunc): Boolean;
    function SearchKey(Key: Pointer; var Index: LongInt): Boolean;
  public
    constructor Create(ArraySize, ADelta: Longint); override;

    function Compare(Key1, Key2: Pointer): LongInt; virtual; abstract;
    function KeyOf(Item: Pointer): Pointer; virtual;
    function Insert(Item: Pointer; KeyOfFunc: TKeyOfFunc;
                    CompareFunc: TCompareFunc): Longint; override;
    function InsertItem(Item: Pointer): Longint; override;
    function IndexOf(Item: Pointer; KeyOfFunc: TKeyOfFunc;
                     CompareFunc: TCompareFunc): LongInt; override;
    property Duplicates: Boolean read FDuplicates write FDuplicates default False;
    property Sorted: Boolean read FSorted write FSorted default True;
  end;

procedure FreeClass(var AClass);

implementation

procedure FreeClass(var AClass);
begin
  if (TObject(AClass) <> nil) then
  begin
    TObject(AClass).Free;
    TObject(AClass) := nil;
  end;
end;

constructor ELongFreeError.Create;
begin
  inherited Create('Couldn''t free memory block.');
end;

constructor ELongRangeError.Create;
begin
  inherited Create('Index not within established range.');
end;

constructor ELongRealloc.Create;
begin
  inherited Create('Couldn''t (re)alloc memory block.');
end;

constructor ELongRealloc.CreateInfo(ExceptInfo: String);
begin
  inherited Create(ExceptInfo);
end;

constructor ELongOverflow.Create;
begin
  inherited CreateInfo('Cannot grow. Collection limit reached.');
end;

constructor EDuplicates.Create;
begin
  inherited Create('Duplicates not allowed.');
end;

function GetMaxSize(NewSize: LongInt): LongInt;
const
  MaxLongCollectionSize: LongInt = MaxLongInt div SizeOf(LongInt);
begin
  if NewSize > MaxLongCollectionSize then
    Result := MaxLongCollectionSize
  else
    Result := NewSize;
end;

{ TSafeCollection }
procedure TSafeCollection.EnterCS;
begin
  EnterCriticalSection(FLock);
end;

procedure TSafeCollection.LeaveCS;
begin
  LeaveCriticalSection(FLock);
end;

procedure TSafeCollection.CheckNotify(Item: Pointer; Action: TCollectionNotify);
begin
  if Item <> nil then
    Notify(Item, Action);
  if Action = cnFree then
    FreeItem(Item);
end;

procedure TSafeCollection.Notify(Item: Pointer; Action: TCollectionNotify);
begin
end;

procedure TSafeCollection.DisposeItem(Item: Pointer);
begin
  CheckNotify(Item, cnFree);
end;

constructor TSafeCollection.Create(ArraySize, ADelta: Longint);
begin
  inherited Create;
  InitializeCriticalSection(FLock);
  FCount  := 0;
  FCurr := HINDEX_ERROR;
  FDelta  := ADelta;
  FMem    := nil;
  FForwrd := True;
  Size    := ArraySize;
  FUserCanAbort := False;
end;

destructor TSafeCollection.Destroy;
begin
  Clear;
  DeleteCriticalSection(FLock);
  inherited Destroy;
end;

function TSafeCollection.CheckIndex(Index: LongInt): Boolean;
begin
  Result := not ((Index < 0) or (Index >= FCount));
end;

function TSafeCollection.CheckSize(Value: LongInt): Boolean;
begin
  Result := not ((Value < 0) or (Value >= FSize));
end;

function TSafeCollection.At(Index: LongInt): Pointer;
begin
  if not CheckIndex(Index) then
    raise ELongRangeError.Create;
  EnterCS;
  try
    Result := Pointer(GetPtr(Index)^);
    CheckNotify(Result, cnAccessed);
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.IndexOf(Item: Pointer; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): LongInt;
begin
  EnterCS;
  try
    if Assigned(Item) then
      for Result := 0 to FCount - 1 do
        if Item = Items[Result] then
          Exit;
    Result := HINDEX_ERROR;
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.IndexOfItem(Item: Pointer): LongInt;
begin
  Result := IndexOf(Item, nil, nil);
end;

function TSafeCollection.Nth(Index : LongInt): Pointer;
begin
  Result := At(Index);
end;

function TSafeCollection.Add(Item: Pointer): Longint;
begin
  EnterCS;
  try
    Result := AtInsert(Count, Item);
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.Append(Item: Pointer): Longint;
begin
  EnterCS;
  try
    Result := AtInsert(Count, Item);
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.Insert(Item: Pointer; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): LongInt;
begin
  Result := Add(Item);
end;

function TSafeCollection.InsertItem(Item: Pointer): Longint;
begin
  Result := Add(Item);
end;

function TSafeCollection.AtInsert(Index: LongInt; Item: Pointer): LongInt;
begin
  Result := Index;
  if FSize = FCount then
  try
    if FDelta = 0 then
      raise ELongOverflow.Create;
    Size := FSize + FDelta;
  except
    on E: ELongRealloc do
    begin
      FreeItem(Item); // no need to notify
      raise;
    end;
  end;
  if FCount > Index then
    MoveItems(Index, Index+1);
  GetPtr(Index)^ := LongInt(Item);
  CheckNotify(Item, cnAdded);
  Inc(FCount);
end;

function TSafeCollection._UNSAFE_AtInsert(Index: Longint; Item: Pointer): Longint;
begin
  Result := AtInsert(Index, Item);
end;

procedure TSafeCollection.FreeAllItems;
var
  I: LongInt;
begin
  if (FCount = 0) then Exit;
  EnterCS;
  try
    for I := 0 to FCount - 1 do
      DisposeItem(Pointer(GetPtr(I)^));
    FCount := 0;
    Size := 0;
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.Clear;
begin
  FreeAllItems;
end;

procedure TSafeCollection.FreeAtIndex(Index: LongInt);
begin
  EnterCS;
  try
    if not CheckIndex(Index) then
      Raise ELongRangeError.Create;
    DisposeItem(Pointer(GetPtr(Index)^));
    if FCount > Index then
      MoveItems(Index+1, Index);
    Dec(FCount);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.FreeThis(Item: Pointer);
var
  Index: LongInt;
begin
  if Item = nil then
    Exit;
  EnterCS;
  try
    Index := IndexOfItem(Item);
    if Index <> HINDEX_ERROR then
      FreeAtIndex(Index);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.FreeItem(Item: Pointer);
begin
  FreeClass(Item);
end;

procedure TSafeCollection.Delete(Index: Longint);
begin
  if not CheckIndex(Index) then
    raise ELongRangeError.Create;
  EnterCS;
  try
    CheckNotify(Pointer(GetPtr(Index)^), cnDeleted);
    MoveItems(Index+1, Index);
    Dec(FCount);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.DeleteItem(Item: Pointer);
var
  Index: LongInt;
begin
  EnterCS;
  try
    Index := IndexOfItem(Item);
    if Index <> HINDEX_ERROR then
      Delete(Index);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.Remove(Item: Pointer);
begin
  DeleteItem(Item);
end;

procedure TSafeCollection.DeleteAll;
var
  I: LongInt;
begin
  if FCount = 0 then
    Exit;
  EnterCS;
  try
    for I := 0 to FCount - 1 do
      CheckNotify(Pointer(GetPtr(I)^), cnDeleted);
    Size := 0;
    FCount := 0;
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.First: Pointer;
begin
  if Count = 0 then Result := nil
  else Result := Items[0];
end;

function TSafeCollection.Last: Pointer;
begin
  if Count = 0 then Result := nil
  else Result := Items[Count-1];
end;

function TSafeCollection.Head: Pointer;
begin
  Result := First;
end;

function TSafeCollection.Tail: Pointer;
begin
  Result := Last;
end;

procedure TSafeCollection.PlaceBefore(aItem, P: Pointer);
var
  I: Integer;
begin
  EnterCS;
  try
    if ( aItem = nil ) or ( P = nil ) then Exit;
    I := IndexOfItem(P);
    AtInsert(I, aItem);
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.FirstThat(Test: Pointer): Pointer;
label
  RETURN;
var
  Index: LongInt;
  Item: Pointer;
  OldForwrd: Boolean;
  callerBP: Cardinal;
begin
  asm
    mov EAX,[EBP]
    mov callerBP,EAX
  end;

  Result := nil;
  if FCount = 0 then Exit;
  OldForwrd := FForwrd;
  FForwrd := True;
  FCurr := 0;
  Index := FCurr;

  repeat
    Item := Items[Index];
    asm
        PUSH    Self
        PUSH    Index
	PUSH    Item
        PUSH    callerBP
        CALL	Test
        POP     ECX
        POP	Item
        POP	Index
        POP	Self
        OR      AL,AL
	JNE     RETURN
    end;
    Inc(Index);
  until (Index = FCount);
RETURN:
  if (Index < FCount) then
  begin
    Result := Item;
    FCurr := Index;
  end
  else
    FCurr := HINDEX_ERROR;
  FForwrd := OldForwrd;
end;

function TSafeCollection.NextThat(Test: Pointer): Pointer;
label
  RETURN;
var
  Index: LongInt;
  Item: Pointer;
  callerBP: Cardinal;
begin
  asm
    mov EAX,[EBP]
    mov callerBP,EAX
  end;
  Result := nil;
  if FCount = 0 then Exit;
  if FForwrd then
    if FCurr < FCount - 1 then Inc(FCurr)
    else Exit
  else
    if FCurr > 0 then Dec(FCurr)
    else Exit;
  Index := FCurr;

  repeat
    Item := Items[Index];
    asm
        PUSH    Self
        PUSH    Index
        PUSH    Item
        PUSH    callerBP
        CALL	Test
        POP     ECX
        POP	Item
        POP	Index
        POP	Self
        OR      AL,AL
	JNE     RETURN
    end;
    if FForwrd then Inc(Index)
    else Dec(Index);
  until (FForwrd and (Index = FCount)) or (not FForwrd and (Index < 0));
RETURN:
  if (FForwrd and (Index <= FCount - 1)) or (not FForwrd and (Index >= 0)) then
  begin
    Result := Item;
    FCurr := Index;
  end
  else
    FCurr := HINDEX_ERROR;
end;

function TSafeCollection.LastThat(Test: Pointer): Pointer;
label
  RETURN;
var
  Index: LongInt;
  Item: Pointer;
  OldForwrd: Boolean;
  callerBP: Cardinal;
begin
  asm
    mov EAX,[EBP]
    mov callerBP,EAX
  end;
  Result := nil;
  if FCount = 0 then Exit;
  FCurr := FCount - 1;
  OldForwrd := FForwrd;
  FForwrd := False;
  Index := FCurr;

  repeat
    Item := Items[Index];
    asm
        PUSH    Self
        PUSH    Index
	PUSH    Item
        PUSH    callerBP
        CALL	Test
        POP     ECX
        POP	Item
	POP	Index
        POP	Self
        OR      AL,AL
	JNE     RETURN
    end;
    Dec(Index);
  until (Index < 0);
RETURN:
  if (Index >= 0) then
  begin
    Result := Item;
    FCurr := Index;
  end
  else
    FCurr := HINDEX_ERROR;
  FForwrd := OldForwrd;
end;

procedure TSafeCollection.ForEach(Action: Pointer);
var
  Index: LongInt;
  Item: Pointer;
  callerBP: Cardinal;
begin
  asm
    mov EAX,[EBP]
    mov callerBP,EAX
  end;
  if FCount = 0 then Exit;
  if FForwrd then Index := 0
  else Index := FCount - 1;
  repeat
    FCurr := Index;
    Item := Items[Index];
    asm
        PUSH    Self
        PUSH    Index
        PUSH	Item
        PUSH    callerBP
        CALL	Action
	POP	ECX
        POP	Item
        POP	Index
        POP	Self
    end;
    if FForwrd then Inc(Index)
    else Dec(Index);
  until (FForwrd and (Index = FCount)) or
        (not FForwrd and (Index < 0)) or
        (FUserCanAbort and (FCurr = -1));
end;

{Called only within Exchange and Slide. DO NOT call this directly,
or make sure indexes are valid.}
procedure TSafeCollection.Swap(Index1, Index2: LongInt);
var
  Item: LongInt;
  P1, P2: PLongInt;
begin
  P1 := GetPtr(Index1);
  P2 := GetPtr(Index2);
  Item := P1^;
  P1^ := P2^;
  P2^ := Item;
end;

procedure TSafeCollection._Exchange(Index1, Index2: LongInt);
begin
  if not CheckIndex(Index1) or not CheckIndex(Index2) then
    Raise ELongRangeError.Create;
  Swap(Index1, Index2);
end;

procedure TSafeCollection.Exchange(Index1, Index2: LongInt);
begin
  EnterCS;
  try
    _Exchange(Index1, Index2);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection._Slide(Index1, Index2: LongInt);
var
  I: LongInt;
begin
  if not CheckIndex(Index1) or not CheckIndex(Index2) then
    Raise ELongRangeError.Create;
  if Index1 < Index2 then
    for I := Index1 to Index2 - 1 do
      Swap(I, I+1)
  else
    if Index1 > Index2 then
      for I := Index1 downto Index2 + 1 do
        Swap(I, I-1);
end;

procedure TSafeCollection.Slide(Index1, Index2: LongInt);
begin
  EnterCS;
  try
    _Slide(Index1, Index2);
  finally
    LeaveCS;
  end;
end;

function TSafeCollection.GetPtr(Index: LongInt): PLongInt;
var
  HMem: THandle;
begin
  HMem := GlobalHandle(FMem);
  GlobalUnlock(HMem);
  Result := Pointer(
              LongInt(GlobalLock(HMem)) + Index * SizeOf(LongInt)
            );
end;

procedure TSafeCollection.MoveItems(Src, Dest: Longint);
begin
  Move(GetPtr(Src)^, GetPtr(Dest)^, (FCount-Src) * SizeOf(Pointer));
end;

procedure TSafeCollection.SetItem(Index: LongInt; Item: Pointer);
begin
  if not CheckSize(Index) then
    raise ELongRangeError.Create;
  EnterCS;
  try
    if FCount > 0 then
      DisposeItem(Items[Index]);
    GetPtr(Index)^ := LongInt(Item);
    CheckNotify(Item, cnAdded);
  finally
    LeaveCS;
  end;
end;

procedure TSafeCollection.SetSize(NewSize: LongInt);
// The GlobalAlloc and LocalAlloc functions are limited to a combined total
// of 65,536 handles for GMEM_MOVEABLE and LMEM_MOVEABLE memory per process.
// This limitation does not apply to GMEM_FIXED or LMEM_FIXED memory.
// If this function succeeds, it allocates at least the amount of memory
// requested. If the actual amount allocated is greater than the amount
// requested, the process can use the entire amount. To determine the actual
// number of bytes allocated, use the GlobalSize function.
var
  AllocBytes: LongInt;
begin
  if NewSize <> FSize then
  begin
    if NewSize = 0 then
    begin
      if GlobalFreePtr(FMem) <> 0 then
        raise ELongFreeError.Create;
      FMem := nil;
      FSize := 0;
    end
    else
    begin
      AllocBytes := GetMaxSize(NewSize) * SizeOf(LongInt);
      if FSize = 0 then FMem := GlobalAllocPtr(GHND, AllocBytes)
      else FMem := GlobalReAllocPtr(FMem, AllocBytes, GHND);
      if FMem = nil then
        raise ELongRealloc.Create;
      FSize := GlobalSize(GlobalHandle(FMem)) div SizeOf(LongInt);
    end;
  end;
end;

{ TSafeSortedCollection }
constructor TSafeSortedCollection.Create(ArraySize, ADelta: LongInt);
begin
  inherited Create(ArraySize, ADelta);
  FDuplicates := False;
  FSorted     := True;
end;

function TSafeSortedCollection.KeyOf(Item: Pointer): Pointer;
begin
  Result := Item;
end;

function TSafeSortedCollection.IndexOf(Item: Pointer; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): LongInt;
var
  I: LongInt;
begin
  EnterCS;
  try
    Result := HINDEX_ERROR;
    if not FSorted then
      Result := Inherited IndexOf(Item, nil, nil)
    else
      if (Assigned(KeyOfFunc) and Assigned(CompareFunc) and
          _UNSAFE_Search(KeyOfFunc(Item), I, KeyOfFunc, CompareFunc)) or
         ((not Assigned(KeyOfFunc) or not Assigned(CompareFunc)) and
          _UNSAFE_Search(KeyOf(Item), I, nil, nil)) then
      begin
        if FDuplicates then
          while (I < Count) and (Item <> Items[I]) do
            Inc(I);
        if I < Count then Result := I;
      end;
  finally
    LeaveCS;
  end;
end;

function TSafeSortedCollection.InsertItem(Item: Pointer): Longint;
begin
  Result := Insert(Item, nil, nil);
end;

function TSafeSortedCollection.Insert(Item: Pointer; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): LongInt;
var
  Ok: Boolean;
begin
  EnterCS;
  try
    Ok := not Sorted or
          (    (Assigned(KeyOfFunc) and Assigned(CompareFunc) and
                not _UNSAFE_Search(KeyOfFunc(Item), Result, KeyOfFunc, CompareFunc)) or
              ((not Assigned(KeyOfFunc) or not Assigned(CompareFunc)) and
                not _UNSAFE_Search(KeyOf(Item), Result, nil, nil))
          );
    if Ok then
      if not Sorted then Result := FCount
      else // Result holds the index
    else // sorted and found a match
      if not FDuplicates then
      begin
        // collection doesn't allow for duplicates, so, the input item
        // has to be disposed. We'll do it or ... you'll do it if
        // you remove next line !!!
        FreeItem(Item); // no need to notify
        raise EDuplicates.Create;
      end;
    AtInsert(Result, Item);
  finally
    LeaveCS;
  end;
end;

function TSafeSortedCollection._UNSAFE_Search(Key: Pointer;
  var Index: LongInt; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): Boolean;
var
  L, H, I, C: LongInt;
  Item: Pointer;
begin
  Result := False;
  L := 0;
  H := FCount - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    Item := Pointer(GetPtr(I)^);
    if Assigned(KeyOfFunc) and Assigned(CompareFunc) then
      C := CompareFunc(KeyOfFunc(Item), Key)
    else
      C := Compare(KeyOf(Item), Key);
    if C < 0 then L := I + 1
    else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        if not FDuplicates then L := I;
      end;
    end;
  end;
  Index := L;
end;

function TSafeSortedCollection._UNSAFE_SearchKey(Key: Pointer;
  var Index: LongInt): Boolean;
begin
  Result := _UNSAFE_Search(Key, Index, nil, nil);
end;

function TSafeSortedCollection.Search(Key: Pointer;
  var Index: LongInt; KeyOfFunc: TKeyOfFunc;
  CompareFunc: TCompareFunc): Boolean;
begin
  EnterCS;
  try
    Result := _UNSAFE_Search(Key, Index, KeyOfFunc, CompareFunc);
  finally
    LeaveCS;
  end;
end;

function TSafeSortedCollection.SearchKey(Key: Pointer; var Index: LongInt): Boolean;
begin
  Result := Search(Key, Index, nil, nil);
end;

end.

